class NotFound(Exception):
    def __init__(self):
        Exception.__init__(self,"404 Not Found")
        
        
class ServerError(Exception):
    def __init__(self):
        Exception.__init__(self,"Server error, try again later")
        

class RateLimit(Exception):
    def __init__(self, headers):
        messageToAdd = ""
        if "Retry-After" in headers:
            self.timeToWait = headers["Retry-After"]
            messageToAdd = ", Retry-After = "+str(self.timeToWait)
        else:
            self.timeToWait = 1
        Exception.__init__(self,"Rate limit exceeded" + messageToAdd)
        
    def waitFor(self):
        return int(self.timeToWait)


class Forbidden(Exception):
    def __init__(self):
        Exception.__init__(self,"Access forbidden, your key may be expired or blacklisted. Check the dev portal : https://developer.riotgames.com/")


class Unauthorized(Exception):
    def __init__(self):
        Exception.__init__(self,"Unauthorized, check your API key")


class BadRequest(Exception):
    def __init__(self):
        Exception.__init__(self,"Bad request, the parameters you give might not be handled by the API. Check the API docs : https://developer.riotgames.com/api-methods/")


class Timeout(Exception):
    def __init__(self):
        Exception.__init__(self,"Timeout reached, try again alter")


from .config import *

def test_status():
    try:
        data = loop.run_until_complete(panth.getStatus())
    except Exception as e:
        print(e)
    
    assert data["region_tag"] == "eu"
    assert data["slug"] == "euw"
    assert type(data["services"]) == list


from .config import *

def test_champion():
    try:
        data = loop.run_until_complete(panth.getChampionRotations())
    except Exception as e:
        print(e)
    
    assert "freeChampionIds" in data
    assert "freeChampionIdsForNewPlayers" in data
    assert "maxNewPlayerLevel" in data
    
    assert type(data["freeChampionIds"]) == list
    assert type(data["freeChampionIdsForNewPlayers"]) == list
    
    


import asyncio, time, datetime

class RateLimiter:
    
    def __init__(self, debug, limits : (int,int) = (10,10), name: str =""):
        
        #Init Async lock for the limiter
        self.lock = asyncio.Lock()
        self.backLock = asyncio.Lock()
        
        #Limits params
        self.limit = limits[0]
        self.duration = limits[1]
        
        #Count and begin of the time window
        self.count = 0
        self.time = 0
        
        #Name of the limiter (for debug prupose)
        self.name = name
        
        # "ID" of the time window
        self.num = 0
        
        #Init outgoing requests counters
        self.currentlyPending = 0
        self.previouslyPending = 0
        
        #Synchronicity state with server time window
        self.synced = False
        
        #Debug mode
        self.debug = debug
        
        
    #Allows to update the limit number of requests
    def updateLimit(self,limit: int):
        self.limit = limit
        
    #Return the set duration of the time window
    def getDuration(self):
        return self.duration
    
    #Return the set limit number of requests
    def getLimit(self):
        return self.limit
    
    async def _reset(self):
        #Be sure to be out of the time window
        while self.time + self.duration >= int(time.mktime(datetime.datetime.utcnow().timetuple())):
            await asyncio.sleep(1)
        
        #Reseting count and time
        self.time = int(time.mktime(datetime.datetime.utcnow().timetuple()))
        self.count = 0
        
        #Manage the count of pending requests
        self.previouslyPending += self.currentlyPending
        self.currentlyPending = 0
        
        #Incr the num for the new time window
        self.num += 1
        #The window is not synch with server yet
        self.synced = False
    
    #Fired when a request is back
    async def getBack(self, num:int, timestamp:int, limit=None):
        with await self.backLock:
            
            #If the current time window is up to date
            if self.time + self.duration > timestamp:
                
                #Decrease the pending counter depending on the time window the request was counted
                if self.num == num:
                    self.currentlyPending -= 1
                else:
                    self.previouslyPending -= 1
                    self.count += 1
                
                #Sync the beginning of time window with server and update the limit
                if not self.synced:
                    if not limit is None:
                        self.updateLimit(limit)
                    self.synced = True
                    self.time = timestamp
                    
            #If the time window is out of date
            else:
                await self._reset()
                if not limit is None:
                    self.updateLimit(limit)
                self.previouslyPending -= 1
                self.count += 1
                self.synced = True
                self.time = timestamp
                

    async def getToken(self):

        with await self.lock:
                #Check if outside time window, reset count
                if self.time + self.duration < time.mktime(datetime.datetime.utcnow().timetuple()):

                    #Wait until it's safe to request and open a new time window
                    while self.previouslyPending >= self.limit:
                        await asyncio.sleep(0.5)
                        
                #If in time window, check count
                elif self.count < self.limit:

                    #Wait until it's safe to request or time window is passed
                    while (self.previouslyPending + self.count) >= self.limit and self.time + self.duration > time.mktime(datetime.datetime.utcnow().timetuple()):
                        await asyncio.sleep(0.5)
                    
                #If count limit reached, await for the end of time window
                else:
                    if self.debug:
                        print(self.name+" limit reached, sleeping for "+str( int((self.time + self.duration) - time.mktime(datetime.datetime.utcnow().timetuple())) + 1)+" seconds")
                        print("Limit : "+str(self.limit)+" per "+str(self.duration)+" / Count : "+ str(self.count))
                    #Await for the next time window
                    await asyncio.sleep( int((self.time + self.duration) - time.mktime(datetime.datetime.utcnow().timetuple())) + 1)

                    
                with await self.backLock:
                    #Double check if a reset has not occured in the mean time
                    timestamp = time.mktime(datetime.datetime.utcnow().timetuple())
                    if self.time + self.duration < timestamp:
                        await self._reset()
                        
                    #Incr count
                    self.count +=1
                    self.currentlyPending += 1
                    
                    return self.num


import asyncio
import aiohttp
import json
from functools import wraps

from .utils import utils as utils
from .utils import exceptions as exc

from .RateLimit.RateLimiterManager import RateLimiterManager

class Pantheon():
    
    BASE_URL = "https://{server}.api.riotgames.com/lol/"
    BASE_URL_TFT = "https://{server}.api.riotgames.com/tft/"
    
    tft_server = {
        "br1":"americas",
        "eun1":"europe",
        "euw1":"europe",
        "jp1":"asia",
        "kr":"asia",
        "la1":"americas",
        "la2":"americas",
        "na1":"americas",
        "oc1":"americas",
        "tr1":"europe",
        "ru":"europe",
        "americas":"americas",
        "europe":"europe",
        "asia":"asia"
    }
    
    def __init__(self, server, api_key, errorHandling = False, requestsLoggingFunction = None, debug=False):
        """
        Initialize an instance of Pantheon class
        
        :param string server: The server Pantheon will target for the requests. An instance is intended to only call one server. Use multiple instances of Pantheon to call multiples servers.
        It can take the values described there : https://developer.riotgames.com/regional-endpoints.html (euw1, na1...)
        :param string api_key: The API key needed to call the Riot API
        :param boolean errorHandling: Precise if Pantheon should autoretry after a ratelimit (429) or server error (5XX). Default is False
        """
        self._key = api_key
        self._server = server
        self._rl = RateLimiterManager(debug)
        
        self.errorHandling = errorHandling
        self.requestsLoggingFunction = requestsLoggingFunction
        self.debug = debug

    def ratelimit(func):
        """
        Decorator for rate limiting.
        It will handle the operations needed by the RateLimiterManager to assure the rate limiting and the change of limits considering the returned header.
        """
        @wraps(func)
        async def waitLimit(*args, **params):
            token = await args[0]._rl.getToken(func.__name__)
            
            response = await func(*args, **params)
            
            try:
                limits = utils.getLimits(response.headers)
                timestamp = utils.getTimestamp(response.headers)
            except:
                limits = None
                timestamp = None
            
            await args[0]._rl.getBack(func.__name__, token, timestamp, limits)
            
            return response
            
        return waitLimit
    
    def errorHandler(func):
        """
        Decorator for handling some errors and retrying if needed.
        """
        @wraps(func)
        async def _errorHandling(*args, **params):
            """
            Error handling function for decorator
            """
            if not args[0].errorHandling:
                return await func(*args, **params)
            else:
                try:
                    return await func(*args, **params)
                #Errors that should be retried
                except exc.RateLimit as e:
                    if args[0].debug:
                        print(e)
                        print("Retrying")
                    i = e.waitFor()
                    while i < 6:
                        await asyncio.sleep(i)
                        try:
                            return await func(*args, **params)
                        except Exception as e2:
                            if args[0].debug:
                                print(e2)
                        i += 2
                    raise e
                except (exc.ServerError, exc.Timeout) as e:
                    if args[0].debug:
                        print(e)
                        print("Retrying")
                    i = 1
                    while i < 6:
                        await asyncio.sleep(i)
                        try:
                            return await func(*args, **params)
                        except (exc.Timeout, exc.ServerError) as e2:
                    
                            pass
                        i += 2
                        if args[0].debug:
                            print(e2)
                            print("Retrying")
                    print("there is no bug")
                    raise e
                except (exc.NotFound, exc.BadRequest) as e:
                    raise e
                except (exc.Forbidden, exc.Unauthorized,) as e:
                    print(e)
                    raise SystemExit(0)
                except Exception as e:
                    raise e
                
        return _errorHandling
    
    def exceptions(func):
        """
        Decorator translating status code into exceptions
        """
        @wraps(func)
        async def _exceptions(*args, **params):
            
            response = await func(*args, **params)
            
            if response is None:
                raise exc.Timeout
            
            elif response.status == 200:
                return json.loads(await response.text())

            elif response.status == 404:
                raise exc.NotFound
                
            elif response.status in [500,502,503,504]:
                raise exc.ServerError
                
            elif response.status == 429:
                raise exc.RateLimit(response.headers)
                
            elif response.status == 403:
                raise exc.Forbidden
                
            elif response.status == 401:
                raise exc.Unauthorized
                
            elif response.status == 400:
                raise exc.BadRequest
                
            elif response.status == 408:
                raise exc.Timeout
                
            else:
                raise Exception("Unidentified error code : "+str(response.status))
            
        return _exceptions
        
    async def fetch(self, url, method="GET", data=None):
        """
        Returns the result of the request of the url given in parameter after attaching the api_key to the header
        """
        
        async with aiohttp.ClientSession() as session:
            headers = {
                "X-Riot-Token": self._key
            }
            
            try:
                if method=="GET":
                    response = await session.request("GET", url, headers=headers)
                else:
                    response = await session.request(method, url, headers=headers, data=json.dumps(data))
            #In case of timeout
            except Exception as e:
                return None
            
            #If a logging function is passed, send it url, status code and headers
            if self.requestsLoggingFunction:
                self.requestsLoggingFunction(url, response.status, response.headers)
            
            #await response.text() needed here in the client session, dunno why
            await response.text()
            return response
    
    #Champion mastery
    @errorHandler
    @exceptions
    @ratelimit
    async def getChampionMasteries(self, summonerId):
        """
        :param string summonerId: summonerId of the player
        
        Returns the result of https://developer.riotgames.com/api-methods/#champion-mastery-v4/GET_getAllChampionMasteries
        """
        return await self.fetch((self.BASE_URL + "champion-mastery/v4/champion-masteries/by-summoner/{summonerId}").format(server=self._server, summonerId=summonerId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getChampionMasteriesByChampionId(self, summonerId, championId):
        """
            :param string summonerId: summonerId of the player
            :param int championId: id of the champion

            Returns the result of https://developer.riotgames.com/api-methods/#champion-mastery-v4/GET_getChampionMastery
        """
        return await self.fetch((self.BASE_URL + "champion-mastery/v4/champion-masteries/by-summoner/{summonerId}/by-champion/{championId}").format(server=self._server, summonerId=summonerId, championId=championId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getChampionMasteriesScore(self, summonerId):
        """
        :param string summonerId: summonerId of the player
        
        Returns the result of https://developer.riotgames.com/api-methods/#champion-mastery-v4/GET_getChampionMasteryScore
        """
        return await self.fetch((self.BASE_URL + "champion-mastery/v4/scores/by-summoner/{summonerId}").format(server=self._server, summonerId=summonerId))
    
    
    #Champions
    @errorHandler
    @exceptions
    @ratelimit
    async def getChampionRotations(self):
        """
        Returns the result of https://developer.riotgames.com/api-methods/#champion-v3/GET_getChampionInfo
        """
        return await self.fetch((self.BASE_URL + "platform/v3/champion-rotations").format(server=self._server))
    
    
    #League
    @errorHandler
    @exceptions
    @ratelimit
    async def getLeagueById(self, leagueId):
        """
        :param string leagueId: id of the league
        
        Returns the result of https://developer.riotgames.com/api-methods/#league-v4/GET_getLeagueById
        """
        return await self.fetch((self.BASE_URL + "league/v4/leagues/{leagueId}").format(server=self._server, leagueId=leagueId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getLeaguePages(self, queue="RANKED_SOLO_5x5", tier="DIAMOND", division="I", page=1):
        """
        :param string queue: queue to get the page of
        :param string tier: tier to get the page of
        :param string division: division to get the page of
        :param int page: page to get
        
        Returns the result of https://developer.riotgames.com/api-methods/#league-v4/GET_getLeagueEntriesForSummoner
        """
        
        return await self.fetch((self.BASE_URL + "league/v4/entries/{queue}/{tier}/{division}?page={page}").format(server=self._server, queue=queue, tier=tier, division=division, page=page))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getLeaguePosition(self, summonerId):
        """
        :param string summonerId: summonerId of the player
        
        Returns the result of https://developer.riotgames.com/api-methods/#league-v4/GET_getLeagueEntriesForSummoner
        """
        return await self.fetch((self.BASE_URL + "league/v4/entries/by-summoner/{summonerId}").format(server=self._server, summonerId=summonerId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getChallengerLeague(self, queue="RANKED_SOLO_5x5"):
        """
        :param string queue: queue to get the challenger league of
            Values accepted : 
             * RANKED_SOLO_5x5 *(default)*
             * RANKED_FLEX_SR
             * RANKED_FLEX_TT
        
        Returns the result of https://developer.riotgames.com/api-methods/#league-v4/GET_getChallengerLeague
        """
        return await self.fetch((self.BASE_URL + "league/v4/challengerleagues/by-queue/{queue}").format(server=self._server, queue=queue))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getGrandmasterLeague(self, queue="RANKED_SOLO_5x5"):
        """
        :param string queue: queue to get the master league of
            Values accepted : 
             * RANKED_SOLO_5x5 *(default)*
             * RANKED_FLEX_SR
             * RANKED_FLEX_TT
        
        Returns the result of https://developer.riotgames.com/api-methods/#league-v4/GET_getGrandmasterLeague
        """
        return await self.fetch((self.BASE_URL + "league/v4/grandmasterleagues/by-queue/{queue}").format(server=self._server, queue=queue))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getMasterLeague(self, queue="RANKED_SOLO_5x5"):
        """
        :param string queue: queue to get the master league of
            Values accepted : 
             * RANKED_SOLO_5x5 *(default)*
             * RANKED_FLEX_SR
             * RANKED_FLEX_TT
        
        Returns the result of https://developer.riotgames.com/api-methods/#league-v4/GET_getMasterLeague
        """
        return await self.fetch((self.BASE_URL + "league/v4/masterleagues/by-queue/{queue}").format(server=self._server, queue=queue))
    
    
    #Status
    @errorHandler
    @exceptions
    @ratelimit
    async def getStatus(self):
        """
        Returns the result of https://developer.riotgames.com/api-methods/#lol-status-v3/GET_getShardData
        """
        return await self.fetch((self.BASE_URL + "status/v3/shard-data").format(server=self._server))
    
    
    #Match
    @errorHandler
    @exceptions
    @ratelimit
    async def getMatch(self, matchId):
        """
        :param int matchId: matchId of the match, also known as gameId
        
        Returns the result of https://developer.riotgames.com/api-methods/#match-v4/GET_getMatch
        """
        return await self.fetch((self.BASE_URL + "match/v4/matches/{matchId}").format(server=self._server, matchId=matchId))
        
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getTimeline(self, matchId):
        """
        :param int matchId: matchId of the match, also known as gameId
        
        Returns the result of https://developer.riotgames.com/api-methods/#match-v4/GET_getMatchTimeline
        """
        return await self.fetch((self.BASE_URL + "match/v4/timelines/by-match/{matchId}").format(server=self._server, matchId=matchId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getMatchlist(self, accountId, params=None):
        """
        :param string accountId: accountId of the player
        :param object params: all key:value params to add to the request
        
        Returns the result of https://developer.riotgames.com/api-methods/#match-v4/GET_getMatchlist
        """
        return await self.fetch((self.BASE_URL + "match/v4/matchlists/by-account/{accountId}{params}").format(server=self._server, accountId=accountId, params = utils.urlParams(params)))
    
    
    #Spectator
    @errorHandler
    @exceptions
    @ratelimit
    async def getCurrentGame(self, summonerId):
        """
        :param string summonerId: summonerId of the player
        
        Returns the result of https://developer.riotgames.com/api-methods/#spectator-v4/GET_getCurrentGameInfoBySummoner
        """
        return await self.fetch((self.BASE_URL + "spectator/v4/active-games/by-summoner/{summonerId}").format(server=self._server, summonerId=summonerId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getFeaturedGame(self):
        """
        Returns the result of https://developer.riotgames.com/api-methods/#spectator-v3/GET_getFeaturedGames
        """
        return await self.fetch((self.BASE_URL + "spectator/v4/featured-games").format(server=self._server))
    
    
    #Summoner
    @errorHandler
    @exceptions
    @ratelimit
    async def getSummoner(self, summonerId):
        """
        :param string summonerId: summonerId of the player
        
        Returns the result of https://developer.riotgames.com/api-methods/#summoner-v4/GET_getBySummonerId
        """
        return await self.fetch((self.BASE_URL + "summoner/v4/summoners/{summonerId}").format(server=self._server, summonerId=summonerId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getSummonerByAccountId(self, accountId):
        """
        :param string accountId: accountId of the player
        
        Returns the result of https://developer.riotgames.com/api-methods/#summoner-v4/GET_getByAccountId
        """
        return await self.fetch((self.BASE_URL + "summoner/v4/summoners/by-account/{accountId}").format(server=self._server, accountId=accountId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getSummonerByName(self, summonerName):
        """
        :param string summonerName: name of the player
        
        Returns the result of https://developer.riotgames.com/api-methods/#summoner-v4/GET_getBySummonerName
        """
        return await self.fetch((self.BASE_URL + "summoner/v4/summoners/by-name/{summonerName}").format(server=self._server, summonerName=summonerName))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getSummonerByPuuId(self, puuId):
        """
        :param string puuId: puuId of the player
        
        Returns the result of https://developer.riotgames.com/apis#summoner-v4/GET_getByPUUID
        """
        return await self.fetch((self.BASE_URL + "summoner/v4/summoners/by-puuid/{puuId}").format(server=self._server, puuId=puuId))
    
    
    #Third Party Code
    @errorHandler
    @exceptions
    @ratelimit
    async def getThirdPartyCode(self, summonerId):
        """
        :param string summonerId: summonerId of the player
        
        Returns the result of https://developer.riotgames.com/api-methods/#third-party-code-v4/GET_getThirdPartyCodeBySummonerId
        """
        return await self.fetch((self.BASE_URL + "platform/v4/third-party-code/by-summoner/{summonerId}").format(server=self._server, summonerId=summonerId))
    
    
    #Tournaments
    @errorHandler
    @exceptions
    @ratelimit
    async def registerProvider(self, region, callback_url, stub=False):
        """
        :param str region: region to get a provider for
        :param str callback_url: url to which a callback will be sent after each match created with a tournament code from this provider
        
        Returns the result of https://developer.riotgames.com/api-methods/#tournament-stub-v4/POST_registerProviderData
        """
        return await self.fetch("https://americas.api.riotgames.com/lol/tournament{stub}/v4/providers".format(stub="-stub" if stub else ""), method="POST", data={"region":region, "url":callback_url})
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def registerTournament(self, providerId, name, stub=False):
        """
        :param int providerId: providerId to create a tournament
        :param str name: name of the tournament
        
        Returns the result of https://developer.riotgames.com/api-methods/#tournament-stub-v4/POST_registerTournament
        """
        return await self.fetch("https://americas.api.riotgames.com/lol/tournament{stub}/v4/tournaments".format(stub="-stub" if stub else ""), method="POST", data={"providerId":providerId, "name":name})
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def createTournamentCode(self, tournamentId, data, nb_codes=1, stub=False):
        """
        :param int tournamentId: tournamentId for which the code will be created
        :param int nb_codes: number of codes to generate
        :param dict data: datafor the code generation, including : 
            list[str] allowedSummonerIds: list of all summonerId (optional)
            str mapType: map for the game
            str pickType: pick type for the game
            str spectatorType: spectator type for the game
            int teamSize: max number of player in a team
            str metadata: additional data to get back with the callback (optional)
        
        Returns the result of https://developer.riotgames.com/api-methods/#tournament-stub-v4/POST_createTournamentCode
        """
        return await self.fetch(("https://americas.api.riotgames.com/lol/tournament{stub}/v4/codes?count={nb_codes}&tournamentId={tournamentId}").format(stub="-stub" if stub else "", tournamentId=tournamentId, nb_codes=nb_codes), method="POST", data=data)
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getLobbyEvents(self, tournamentCode, stub=False):
        """
        :param int providerId: providerId to create a tournament
        :param str name: name of the tournament
        
        Returns the result of https://developer.riotgames.com/api-methods/#tournament-stub-v4/GET_getLobbyEventsByCode
        """
        return await self.fetch("https://americas.api.riotgames.com/lol/tournament{stub}/v4/lobby-events/by-code/{code}".format(stub="-stub" if stub else "", code=tournamentCode))

    # TFT
    @errorHandler
    @exceptions
    @ratelimit
    async def getTFTLeagueById(self, leagueId):
        """
        :param string leagueId: id of the league
        
        Returns the result of https://developer.riotgames.com/apis#tft-league-v1/GET_getLeagueById
        """
        return await self.fetch((self.BASE_URL_TFT + "league/v1/leagues/{leagueId}").format(server=self._server, leagueId=leagueId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getTFTLeaguePages(self,  tier="DIAMOND", division="I", page=1):
        """
        :param string tier: tier to get the page of
        :param string division: division to get the page of
        :param int page: page to get
        
        Returns the result of https://developer.riotgames.com/apis#tft-league-v1/GET_getLeagueEntries
        """
        return await self.fetch((self.BASE_URL_TFT + "league/v1/entries/{tier}/{division}?page={page}").format(server=self._server, tier=tier, division=division, page=page))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getTFTLeaguePosition(self, summonerId):
        """
        :param string summonerId: summonerId of the player
        
        Returns the result of https://developer.riotgames.com/apis#tft-league-v1/GET_getLeagueEntriesForSummoner
        """
        return await self.fetch((self.BASE_URL_TFT + "league/v1/entries/by-summoner/{summonerId}").format(server=self._server, summonerId=summonerId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getTFTChallengerLeague(self):
        """
        Returns the result of https://developer.riotgames.com/apis#tft-league-v1/GET_getChallengerLeague
        """
        return await self.fetch((self.BASE_URL_TFT + "league/v1/challenger").format(server=self._server))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getTFTGrandmasterLeague(self):
        """
        Returns the result of https://developer.riotgames.com/apis#tft-league-v1/GET_getGrandmasterLeague
        """
        return await self.fetch((self.BASE_URL_TFT + "league/v1/grandmaster").format(server=self._server))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getTFTMasterLeague(self):
        """
        Returns the result of https://developer.riotgames.com/apis#tft-league-v1/GET_getMasterLeague
        """
        return await self.fetch((self.BASE_URL_TFT + "league/v1/master").format(server=self._server))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getTFTMatch(self, matchId):
        """
        :param int matchId: matchId of the match, also known as gameId
        
        Returns the result of https://developer.riotgames.com/api-methods/#match-v4/GET_getMatch
        """
        return await self.fetch((self.BASE_URL_TFT + "match/v1/matches/{matchId}").format(server= self.tft_server[self._server], matchId=matchId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getTFTMatchlist(self, puuId):
        """
        :param string puuId: puuId of the player
        
        Returns the result of https://developer.riotgames.com/apis#tft-match-v1/GET_getMatchIdsByPUUID
        """
        return await self.fetch((self.BASE_URL_TFT + "match/v1/matches/by-puuid/{puuId}/ids").format(server= self.tft_server[self._server], puuId=puuId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getTFTSummoner(self, summonerId):
        """
        :param string summonerId: summonerId of the player
        
        Returns the result of https://developer.riotgames.com/apis#tft-summoner-v1/GET_getBySummonerId
        """
        return await self.fetch((self.BASE_URL_TFT + "summoner/v1/summoners/{summonerId}").format(server=self._server, summonerId=summonerId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getTFTSummonerByAccountId(self, accountId):
        """
        :param string accountId: accountId of the player
        
        Returns the result of https://developer.riotgames.com/apis#tft-summoner-v1/GET_getByAccountId
        """
        return await self.fetch((self.BASE_URL_TFT + "summoner/v1/summoners/by-account/{accountId}").format(server=self._server, accountId=accountId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getTFTSummonerByPuuId(self, puuId):
        """
        :param string puuId: puuId of the player
        
        Returns the result of https://developer.riotgames.com/apis#tft-summoner-v1/GET_getByPUUID
        """
        return await self.fetch((self.BASE_URL_TFT + "summoner/v1/summoners/by-puuid/{puuId}").format(server=self._server, puuId=puuId))
    
    
    @errorHandler
    @exceptions
    @ratelimit
    async def getTFTSummonerByName(self, summonerName):
        """
        :param string summonerName: name of the player
        
        Returns the result of https://developer.riotgames.com/apis#tft-summoner-v1/GET_getBySummonerName
        """
        return await self.fetch((self.BASE_URL_TFT + "summoner/v1/summoners/by-name/{summonerName}").format(server=self._server, summonerName=summonerName))
    
    


from .RateLimiter import RateLimiter

class RateLimiterManager:
    
    #Default application rate limit
    defaultApplicationLimits = [(20,1),(100,120)]
    
    defaultMethodsLimits = {
        "getChampions":[(400,60)],
        "getChampionsById":[(400,60)],
        "getChampionMasteries":[(20000,10),(1200000,600)],
        "getLeagueById":[(500,10)],
        "getLeaguePosition":[(35,60)],
        "getChallengerLeague":[(500,10)],
        "getMasterLeague":[(500,10)],
        "getMatch":[(500,10)],
        "getTimeline":[(500,10)],
        "getMatchlist":[(1000,10)],
        "getSummoner":[(600,60)],
        "getSummonerByName":[(600,60)],
        "getSummonerByAccount":[(1000,60)],
    }
    
    def __init__(self, debug):
        
        self.debug=debug
        
        self.application = []
        for appLimit in self.defaultApplicationLimits:
            self.application.append(RateLimiter(self.debug,appLimit, "App"))
            
        self.methods = {}
        
        for method in self.defaultMethodsLimits:
            self.methods[method] = []
            for methodLimit in self.defaultMethodsLimits[method]:
                self.methods[method].append(RateLimiter(self.debug,methodLimit, method))
        
    
    def updateApplicationLimit(self, duration:int, limit:int):
        for appLimit in self.application:
            if duration == appLimit.getDuration():
                appLimit.updateLimit(limit)
                return
        self.application.append(RateLimiter(self.debug,(limit,duration),"App"))
        
    def deleteApplicationLimit(self, duration:int):
        for appLimit in self.application:
            if duration == appLimit.getDuration():
                self.application.remove(appLimit)
                return
    
    def displayApplicationLimit(self):
        for appLimit in self.application:
            print(str(appLimit.getLimit())+" : "+str(appLimit.getDuration()))
            
    
    def displayApplicationLimiters(self):
        for appLimit in self.application:
            appLimit.displayLimiter()
    
    
    def updateMethodsLimit(self, method:str, duration:int, limit:int):
        if method in self.methods:
            for methodLimit in self.methods[method]:
                if duration == methodLimit.getDuration():
                    methodLimit.updateLimit(limit)
                    return
            self.methods[method].append(RateLimiter(self.debug,(limit,duration),method))
        else:
            self.methods[method] = []
            self.methods[method].append(RateLimiter(self.debug,(limit,duration),method))
        
    def deleteMethodsLimit(self, method:str, duration:int):
        for methodLimit in self.methods[method]:
            if duration == methodLimit.getDuration():
                self.methods[method].remove(methodLimit)
                return
            
    async def getBack(self, method:str, token, timestamp:int, limits):
        if limits is None:
            for i,appLimit in enumerate(self.application):
                await appLimit.getBack(token[0][i], timestamp)
            for i,methodLimit in enumerate(self.methods[method]):
                await methodLimit.getBack(token[1][i], timestamp)
        else:
            appLimitsToDelete = []
            methodsLimitsToDelete = []
            
            for i,appLimit in enumerate(self.application):
                if appLimit.getDuration() in limits[0]:
                    await appLimit.getBack(token[0][i], timestamp, limits[0][appLimit.getDuration()])
                    del(limits[0][appLimit.getDuration()])
                else:
                    #If the limit is not in the returned header, consider it out of date hence, queue to delete it
                    appLimitsToDelete.append(appLimit.getDuration())
                    
            #Delete the out of date limits
            for i in appLimitsToDelete:
                self.deleteApplicationLimit(i)
            
            for duration in limits[0]:
                #If the limit exists in the returned header but not in the manager, create it
                self.updateApplicationLimit(duration,limits[0][duration])
                
            
            for i,methodLimit in enumerate(self.methods[method]):
                if methodLimit.getDuration() in limits[1]:
                    await methodLimit.getBack(token[1][i], timestamp, limits[1][methodLimit.getDuration()])
                    del(limits[1][methodLimit.getDuration()])
                else:
                    #If the limit is not in the returned header, consider it out of date hence, queue to delete it
                    methodsLimitsToDelete.append(methodLimit.getDuration())
            
            #Delete the out of date limits
            for i in methodsLimitsToDelete:
                self.deleteMethodsLimit(method, i)
                
            for duration in limits[1]:
                #If the limit exists in the returned header but not in the manager, create it
                self.updateMethodsLimit(method, duration,limits[1][duration])
    
    def displayMethodsLimit(self):
        for method in self.methods:
            print(method)
            for methodLimit in self.methods[method]:
                print(str(methodLimit.getLimit())+" : "+str(methodLimit.getDuration()))
        print()
    
    def displayMethodLimiters(self, method:str):
        for methodLimit in self.methods[method]:
            methodLimit.displayLimiter()
    
    
        
    async def getToken(self, method:str):
        appToken=[]
        methodToken=[]
        for appLimit in self.application:
            appToken.append(await appLimit.getToken())
        
        if not method in self.methods:
            self.updateMethodsLimit(method, 10, 20000)
        
        for methodLimit in self.methods[method]:
            methodToken.append(await methodLimit.getToken())
        return (appToken,methodToken)
            
        


from .config import *


def test_providers():
    try:
        data = loop.run_until_complete(panth.registerProvider(tournament_region, tournament_url, stub))
    except exc.Unauthorized as e:
        pytest.skip("API key unauthorized for tournament")
    except Exception as e:
        print(e)
    
    assert type(data) == int

def test_tournaments():
    try:
        provider_id = loop.run_until_complete(panth.registerProvider(tournament_region, tournament_url, stub))
        data = loop.run_until_complete(panth.registerTournament(provider_id, tournament_name, stub))
    except exc.Unauthorized as e:
        pytest.skip("API key unauthorized for tournament")
    except Exception as e:
        print(e)
    
    assert type(data) == int

    
def test_code():
    data_input = {
        "mapType": "SUMMONERS_RIFT",
        "metadata": "",
        "pickType": "BLIND_PICK",
        "spectatorType": "NONE",
        "teamSize": 5
    }
    try:
        provider_id = loop.run_until_complete(panth.registerProvider(tournament_region, tournament_url, stub))
        tournament_id = loop.run_until_complete(panth.registerTournament(provider_id, tournament_name, stub))
        data = loop.run_until_complete(panth.createTournamentCode(tournament_id, data_input, 1, stub))
    except exc.Unauthorized as e:
        pytest.skip("API key unauthorized for tournament")
    except Exception as e:
        print(e)
    
    assert type(data) == list
    assert len(data) == 1

    
def test_multiple_codes():
    data_input = {
        "mapType": "SUMMONERS_RIFT",
        "metadata": "",
        "pickType": "BLIND_PICK",
        "spectatorType": "NONE",
        "teamSize": 5
    }
    try:
        provider_id = loop.run_until_complete(panth.registerProvider(tournament_region, tournament_url, stub))
        tournament_id = loop.run_until_complete(panth.registerTournament(provider_id, tournament_name, stub))
        data = loop.run_until_complete(panth.createTournamentCode(tournament_id, data_input, 5, stub))
    except exc.Unauthorized as e:
        pytest.skip("API key unauthorized for tournament")
    except Exception as e:
        print(e)
    
    assert type(data) == list
    assert len(data) == 5

    
def test_lobby():
    try:
        provider_id = loop.run_until_complete(panth.registerProvider(tournament_region, tournament_url, stub))
        data = loop.run_until_complete(panth.getLobbyEvents(provider_id, stub))
    except exc.Unauthorized as e:
        pytest.skip("API key unauthorized for tournament")
    except Exception as e:
        print(e)
    
    assert "eventList" in data
    assert type(data["eventList"]) == list


from .config import *

def test_active_game():
    try:
        data = loop.run_until_complete(panth.getCurrentGame(summonerId))
    except exc.NotFound as e:
        pytest.skip("not currently in game")
    except Exception as e:
        print(e)
    
    assert any([p["summonerId"] == summonerId for p in data["participants"]])
    

def test_featured():
    try:
        data = loop.run_until_complete(panth.getFeaturedGame())
    except Exception as e:
        print(e)
        
    assert type(data["gameList"]) == list
    


from .config import *

def test_by_summonerId():
    try:
        data = loop.run_until_complete(panth.getSummoner(summonerId))
    except Exception as e:
        print(e)
    
    assert data["name"] == name
    assert data["id"] == summonerId

    
def test_by_summonerName():
    try:
        data = loop.run_until_complete(panth.getSummonerByName(name))
    except Exception as e:
        print(e)
    
    assert data["name"] == name
    assert data["id"] == summonerId

    
def test_by_accountId():
    try:
        data = loop.run_until_complete(panth.getSummonerByAccountId(accountId))
    except Exception as e:
        print(e)
    
    assert data["name"] == name
    assert data["accountId"] == accountId

    
def test_by_puuId():
    try:
        data = loop.run_until_complete(panth.getSummonerByPuuId(puuid))
    except Exception as e:
        print(e)
    
    assert data["name"] == name
    assert data["puuid"] == puuid


from .config import *

def test_match():
    try:
        data = loop.run_until_complete(panth.getMatch(matchId))
    except Exception as e:
        print(e)
    
    assert data["gameId"] == matchId
    assert "participants" in data

    
def test_timeline():
    try:
        data = loop.run_until_complete(panth.getTimeline(matchId))
    except Exception as e:
        print(e)
    
    assert "frames" in data
    

def test_matchlist():
    try:
        data = loop.run_until_complete(panth.getMatchlist(accountId))
    except exc.NotFound as e:
        pytest.skip("no match found")
    except Exception as e:
        print(e)
    
    assert type(data["matches"]) == list
    

def test_matchlist_params():
    try:
        data = loop.run_until_complete(panth.getMatchlist(accountId, {"queue":420}))
    except exc.NotFound as e:
        pytest.skip("no match found")
    except Exception as e:
        print(e)
    
    assert all([m["queue"] == 420 for m in data["matches"]])
    

def test_matchlist_params_set():
    try:
        data = loop.run_until_complete(panth.getMatchlist(accountId, {"queue":[420,430]}))
    except exc.NotFound as e:
        pytest.skip("no match found")
    except Exception as e:
        print(e)
    
    assert not all([m["queue"] == 420 for m in data["matches"]])
    assert all([m["queue"] == 420 or m["queue"] == 430 for m in data["matches"]])
    

def test_matchlist_params_multi():
    try:
        data = loop.run_until_complete(panth.getMatchlist(accountId, {"queue":420, "champion":89}))
    except exc.NotFound as e:
        pytest.skip("no match found")
    except Exception as e:
        print(e)
    
    assert all([m["queue"] == 420 and m["champion"] == 89 for m in data["matches"]])
    


from .config import *

def test_leagueId():
    try:
        data = loop.run_until_complete(panth.getLeagueById(leagueId))
    except Exception as e:
        print(e)
    
    assert data["leagueId"] == leagueId
    assert type(data["entries"]) == list
    
    
@pytest.mark.skipif(too_early, reason="Too early in the season")
def test_apex_challenger():
    try:
        data = loop.run_until_complete(panth.getChallengerLeague())
    except Exception as e:
        print(e)
    
    assert data["tier"] == "CHALLENGER"
    assert data["queue"] == "RANKED_SOLO_5x5"
    assert type(data["entries"]) == list
    
    
@pytest.mark.skipif(too_early, reason="Too early in the season")
def test_apex_challenger_flex():
    try:
        data = loop.run_until_complete(panth.getChallengerLeague("RANKED_FLEX_SR"))
    except Exception as e:
        print(e)
    
    assert data["tier"] == "CHALLENGER"
    assert data["queue"] == "RANKED_FLEX_SR"
    assert type(data["entries"]) == list
    
    
@pytest.mark.skipif(too_early, reason="Too early in the season")
def test_apex_grandmaster():
    try:
        data = loop.run_until_complete(panth.getGrandmasterLeague())
    except Exception as e:
        print(e)
    
    assert data["tier"] == "GRANDMASTER"
    assert data["queue"] == "RANKED_SOLO_5x5"
    assert type(data["entries"]) == list
    
    
@pytest.mark.skipif(too_early, reason="Too early in the season")
def test_apex_grandmaster_flex():
    try:
        data = loop.run_until_complete(panth.getGrandmasterLeague("RANKED_FLEX_SR"))
    except Exception as e:
        print(e)
    
    assert data["tier"] == "GRANDMASTER"
    assert data["queue"] == "RANKED_FLEX_SR"
    assert type(data["entries"]) == list
    
    
@pytest.mark.skipif(too_early, reason="Too early in the season")
def test_apex_master():
    try:
        data = loop.run_until_complete(panth.getMasterLeague())
    except Exception as e:
        print(e)
    
    assert data["tier"] == "MASTER"
    assert data["queue"] == "RANKED_SOLO_5x5"
    assert type(data["entries"]) == list
    
    
@pytest.mark.skipif(too_early, reason="Too early in the season")
def test_apex_master_flex():
    try:
        data = loop.run_until_complete(panth.getMasterLeague("RANKED_FLEX_SR"))
    except Exception as e:
        print(e)
    
    assert data["tier"] == "MASTER"
    assert data["queue"] == "RANKED_FLEX_SR"
    assert type(data["entries"]) == list

    
def test_league_entries_by_summonerId():
    try:
        data = loop.run_until_complete(panth.getLeaguePosition(summonerId))
    except Exception as e:
        print(e)
    
    assert type(data) == list
    

def test_league_entries():
    try:
        data = loop.run_until_complete(panth.getLeaguePages())
    except Exception as e:
        print(e)
    
    assert type(data) == list
    
    if len(data) > 0:
        entry = data[0]
        assert entry["queueType"] == "RANKED_SOLO_5x5"
        assert entry["tier"] == "DIAMOND"
        assert entry["rank"] == "I"
        
    else:
        pytest.skip("not enough player at this rank")
    

def test_league_entries_params():
    try:
        data = loop.run_until_complete(panth.getLeaguePages(queue="RANKED_FLEX_SR", tier="SILVER", division="III"))
    except Exception as e:
        print(e)
    
    assert type(data) == list
    
    if len(data) > 0:
        entry = data[0]
        assert entry["queueType"] == "RANKED_FLEX_SR"
        assert entry["tier"] == "SILVER"
        assert entry["rank"] == "III"
        
    else:
        pytest.skip("not enough player at this rank")
    

def test_league_entries_pages():
    try:
        data = loop.run_until_complete(panth.getLeaguePages(queue="RANKED_FLEX_SR", tier="SILVER", division="III", page=1))
    except Exception as e:
        print(e)
        
    try:
        data_2 = loop.run_until_complete(panth.getLeaguePages(queue="RANKED_FLEX_SR", tier="SILVER", division="III", page=2))
    except Exception as e:
        print(e)
    
    assert type(data) == list
    
    if len(data) > 0 and len(data_2) > 0:
        entry = data[0]
        assert entry["queueType"] == "RANKED_FLEX_SR"
        assert entry["tier"] == "SILVER"
        assert entry["rank"] == "III"
        
        entry_2 = data_2[0]
        
        assert not entry == entry_2
        
    else:
        pytest.skip("not enough player at this rank")






from .config import *

def test_code():
    try:
        data = loop.run_until_complete(panth.getThirdPartyCode(summonerId))
    except exc.NotFound as e:
        pytest.skip("code not set")
    except Exception as e:
        print(e)
    
    assert type(data) == str


from .config import *

def test_tft_match():
    try:
        data = loop.run_until_complete(panth.getTFTMatch(tft_matchId))
    except Exception as e:
        print(e)
    
    assert "metadata" in data
    assert "info" in data
    assert data["metadata"]["match_id"] == tft_matchId

    
def test_tft_matchlist():
    try:
        data = loop.run_until_complete(panth.getTFTMatchlist(puuid))
    except exc.NotFound as e:
        pytest.skip("no match found")
    except Exception as e:
        print(e)
    
    assert type(data) == list
    


from .config import *


def test_championMastery_by_summonerId():
    try:
        data = loop.run_until_complete(panth.getChampionMasteries(summonerId))
    except Exception as e:
        print(e)
    
    assert type(data) == list
    assert len(data) > 0
    

def test_championMastery_by_summonerId_by_championId():
    try:
        data = loop.run_until_complete(panth.getChampionMasteriesByChampionId(summonerId, 89)) # Leona
    except Exception as e:
        print(e)
    
    assert data["championId"] == 89
    assert data["championLevel"] == 7


def test_championMasteryScore_by_summonerId():
    try:
        data = loop.run_until_complete(panth.getChampionMasteriesScore(summonerId))
    except Exception as e:
        print(e)
    
    assert type(data) == int
    assert data > 200
    
    


import asyncio
import pytest

from pantheon import pantheon
from pantheon.utils import exceptions as exc

# API details
api_key = "RGAPI-b2336aa3-2a88-4ca2-9fc0-dc8388788f43"
server = "euw1"
panth = pantheon.Pantheon(server, api_key, errorHandling=True)

# Summoner details
name = "Canisback"
accountId = "VRTAqAZgHiTwK90Xd9hAaRejrTD0D9n69izLSqQwKf9sug"
summonerId = "Oizbr1dzsLM9ygEADNequ88CIRCq2wDV5EeWcxZQFWSKFvc"
puuid = "hhH5SNlbp-0xYiernub2nxBdMeuNxQRRDcCI4bXh4xoE4bS_II1d-lemvVT3Q2Uf0ZdiJmnr0GOhBg"


# League details
leagueId = "d112cf40-35be-11e9-947f-c81f66db01ef"

## Set to True if skipping apex tiers
too_early = False

# Match details
matchId = 4259542242

# Tournament details
stub = True
tournament_region = "EUW"
tournament_url = "http://test.com"
tournament_name = "Test"

# TFT details
tft_leagueId = "30032c60-f82a-11e9-8d92-a2a060ae885a"

tft_matchId = "EUW1_4268130192"

loop = asyncio.get_event_loop() 




import time, datetime

def getLimits(headers):
    if 'X-Method-Rate-Limit' in headers and 'X-App-Rate-Limit' in headers:
        appLimits = {}
        for appLimit in headers['X-App-Rate-Limit'].split(","):
            appLimits[int(appLimit.split(":")[1])] = int(appLimit.split(":")[0])

        methodLimits = {}
        for methodLimit in headers['X-Method-Rate-Limit'].split(","):
            methodLimits[int(methodLimit.split(":")[1])] = int(methodLimit.split(":")[0])

        return (appLimits,methodLimits)
    return None

def dateToTimestamp(date):
    return int(time.mktime(datetime.datetime.strptime(date, '%a, %d %b %Y %H:%M:%S GMT').timetuple()))

def getTimestamp(headers):
    try:
        timestamp = dateToTimestamp(headers['Date'])
    except Exception as e:
        timestamp = int(time.time())
    return timestamp

def urlParams(params):
    if params is None:
        return ""
    else:
        strParams = "?"
        for i in params:
            if type(params[i]) == list:
                for p in params[i]:
                    strParams+= i+"="+str(p)+"&"
            else:
                strParams+= i+"="+str(params[i])+"&"
        return strParams[:-1]


from setuptools import setup, find_packages

from os import path
this_directory = path.dirname(__file__)
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()



setup(
  name = 'pantheon',
  packages = find_packages(),
  version = '1.0.11',
  description = 'Riot API library for Python and asyncio',
  long_description=long_description,
  long_description_content_type='text/markdown',
  author = 'Canisback',
  author_email = 'canisback@gmail.com',
  url = 'https://github.com/Canisback/pantheon',
  keywords = ['Riot API', 'python'],
  classifiers = [],
  install_requires=[
    "asyncio",
    "aiohttp"
  ],
)
